
#include "bcurv.h"



#define New(TYPE)             ((TYPE *) malloc(sizeof(TYPE)))
#define New_array(NELT, TYPE) ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))
#define Realloc_array(PTR, NEW_SIZE, TYPE) \
                    ((TYPE *) realloc(PTR, (size_t)(NEW_SIZE * sizeof(TYPE))))
#define Free(P)               free(P)


struct b_crv_t;

typedef void (*interpolate_method_t)(double param,
                                     b_curve_t curve,
                                     double *x, double *y);

typedef struct b_curve_methods_t {
  interpolate_method_t  interpolate;
}              b_curve_methods_t;


typedef int bool_t;

typedef void *b_curve_data_t;




typedef struct b_crv_t {
  b_curve_type_enum    type;
  b_curve_methods_t   *methods;
  b_curve_data_t       data;
}              b_crv_t;

typedef struct location_t {
  double x, y;
}              location_t;



static void 
interpolate_line(double eta, b_curve_t c, double *x, double *y);

static b_curve_methods_t line_methods = {
  interpolate_line
};

static void 
interpolate_circ_arc(double eta, b_curve_t c, double *x, double *y);

static b_curve_methods_t circ_arc_methods = {
  interpolate_circ_arc
};



static double 
line_length(double fx, double fy, double lx, double ly);
static double 
line_length(double fx, double fy, double lx, double ly)
{
  double temp1, temp2;
  temp1 = (double)(fx - lx);  temp1 *= temp1;
  temp2 = (double)(fy - ly);  temp2 *= temp2;
  return sqrt(temp1 + temp2); 
}




double 
bcurv_find_close_eta(b_curve_t curve, double x, double y,
                     double eta_lower, double eta_upper)
{
  b_crv_t *c = (b_crv_t *)curve;
  double step;
  double e, cx, cy, dist, d, reteta;

  step   = ((eta_upper-eta_lower)/100);
  reteta = e = eta_lower;
  ((*c->methods).interpolate)(e, c, &cx, &cy);
  dist = line_length(x, y, cx, cy);
  for (e = eta_lower+step; e <= eta_upper; e += step) {
    ((*c->methods).interpolate)(e, c, &cx, &cy);
    if ((d = line_length(x, y, cx, cy)) < dist) {
      dist   = d;
      reteta = e;
    } else 
      return reteta;
  }
  return reteta;
}




void
bcurv_interpolate(double param,
                  b_curve_t curve,
                  double *x, double *y)
{
  b_crv_t *c = (b_crv_t *)curve;

  (*c->methods->interpolate)(param, curve, x, y);
}




/* STRAIGHT LINE SEGMENT */

typedef struct line_data_t {
  location_t   first_point;
  location_t   last_point;
}              line_data_t;


b_curve_t 
bcurv_make_line(double fx, double fy, double lx, double ly)
{
  b_crv_t *c;
  line_data_t *d;

  c = New(b_crv_t);
  if (c == NULL) return NULL;
  
  c->methods = &line_methods;
  c->type    = STRAIGHT_LINE_SEGMENT;
  
  d = New(line_data_t);
  if (d == NULL) return NULL;

  d->first_point.x = fx;
  d->first_point.y  = fy;
  d->last_point.x   = lx;
  d->last_point.y   = ly;
  c->data    = (b_curve_data_t)d;

  return (b_curve_t)c;
}


static void 
interpolate_line(double eta, b_curve_t curve, double *x, double *y)
{
  b_crv_t *c = (b_crv_t *)curve;
  line_data_t *d = (line_data_t *)c->data;
  
  *x = (1 - eta) * d->first_point.x + eta * d->last_point.x;
  *y = (1 - eta) * d->first_point.y + eta * d->last_point.y;
}



/* CIRCULAR ARC */

typedef struct circ_arc_data_t {
  location_t           center_point;
  location_t           first_point;
  location_t           last_point;
  double               radius;
  double               start_angle;
  double               end_angle;
  double               internal_angle;
  bool_t               in_opposite_dir;
}              circ_arc_data_t;



b_curve_t 
bcurv_make_circ_arc(double fx, double fy, /* first point */
                    double lx, double ly, /* last point */
                    int intermed_p_is_valid, /* Is the following valid? */
                    double ix, double iy, /* intermediate point */
                    double cx, double cy, /* center */
                    int opposite_dir /* Default orientation is ccw; */
                                     /* if this parameter is set, the */
                                     /* orientation is cw */
                    )
{
  b_crv_t *c;
  double temp1, temp2, vecx, vecy, vecl, proj1, proj2, ia;
  double true_cx, true_cy, true_radius, true_start_ang, true_end_ang;
  circ_arc_data_t *d;

  c = New(b_crv_t);
  if (c == NULL) return NULL;
  
  c->methods = &circ_arc_methods;
  c->type  = CIRCULAR_ARC;
 
  d = New(circ_arc_data_t);
  if (d == NULL) return NULL;

  d->first_point.x      = fx;
  d->first_point.y      = fy;
  d->last_point.x       = lx;
  d->last_point.y       = ly;

  true_cx = cx; true_cy = cy;
  temp1 = fx-cx;  temp1 *= temp1;
  temp2 = fy-cy;  temp2 *= temp2;
  true_radius = sqrt(temp1 + temp2);
#define DOT_P_2(v1x, v1y, v2x, v2y) (((v1x)*(v2x))+((v1y)*(v2y)))
#define ACCURACY_EPS    1.e-7
#define VERY_SMALL_NMBR 1.e-30
  vecx  = lx - fx; vecy = ly - fy;
  temp1 = vecx;  temp1 *= temp1;
  temp2 = vecy;  temp2 *= temp2;
  vecl  = sqrt(temp1 + temp2);
  vecx /= vecl; vecy /= vecl;	/* unit vector */
  proj1 = DOT_P_2(cx-fx, cy-fy, vecx, vecy);
  proj2 = DOT_P_2(cx-lx, cy-ly, -vecx, -vecy);
  if (fabs(proj1-proj2) > ACCURACY_EPS) {
    true_cx = cx - 0.5 * (proj1-proj2)*vecx;
    true_cy = cy - 0.5 * (proj1-proj2)*vecy;
    temp1 = fx-true_cx;  temp1 *= temp1;
    temp2 = fy-true_cy;  temp2 *= temp2;
    true_radius = sqrt(temp1 + temp2);
  }
  vecx = fx - true_cx;
  vecy = fy - true_cy; 
  if (fabs(vecx) < VERY_SMALL_NMBR)    {
    if (vecy < 0.)
      true_start_ang = 1.5 * M_PI;
    else
      true_start_ang = 0.5 * M_PI;
  }  else    {
    true_start_ang = atan(fabs(vecy / vecx));
    if (vecx > 0.) {
      if (vecy > 0.)
	true_start_ang = true_start_ang;
      else
	true_start_ang = -true_start_ang;
    }      else	{
      if (vecy > 0.) 
	true_start_ang = M_PI - true_start_ang;
      else
	true_start_ang = M_PI + true_start_ang;
    }
  }
  if (true_start_ang < 0)
    true_start_ang += 2. * M_PI;
  vecx = lx - true_cx;
  vecy = ly - true_cy;
  if (fabs(vecx) < VERY_SMALL_NMBR)    {
    if (vecy < 0.)
      true_end_ang = 1.5 * M_PI;
    else
      true_end_ang = 0.5 * M_PI;
  }  else    {
    true_end_ang = atan(fabs(vecy / vecx));
    if (vecx > 0.) {
      if (vecy > 0.)
	true_end_ang = true_end_ang;
      else
	true_end_ang = -true_end_ang;
    }      else	{
      if (vecy > 0.) 
	true_end_ang = M_PI - true_end_ang;
      else
	true_end_ang = M_PI + true_end_ang;
    }
  }
  if (true_end_ang < 0)
    true_end_ang += 2. * M_PI;

  d->center_point.x     = true_cx;
  d->center_point.y     = true_cy;
  d->radius             = true_radius;
  d->start_angle        = true_start_ang;
  d->end_angle          = true_end_ang;

  ia = true_end_ang - true_start_ang;
  if (opposite_dir) {
    if (ia > 0)
      ia = ia - 2. * M_PI;
    else
      ia = ia + 2. * M_PI;
  }
  d->internal_angle = ia;

  c->data    = (b_curve_data_t)d;

  return (b_curve_t)c;
}




static void 
interpolate_circ_arc(double eta, b_curve_t curve, double *x, double *y)
{
  b_crv_t *c = (b_crv_t *)curve;
  circ_arc_data_t *d = (circ_arc_data_t *)c->data;
  double a, r;

  a  = eta * d->internal_angle + d->start_angle;
  r  = d->radius;
  *x = d->center_point.x + r * cos(a);
  *y = d->center_point.y + r * sin(a);
}


