

#include "hexmesh.h"
#include "cracksurf.h"


#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif


typedef struct triangle_obj_t {
  int       id;
  WCRec     pts[3];
  int       vertices[3];
}              triangle_obj_t;


typedef struct vertex_obj_t {
  int    id;
  WCRec  loc;
  int    is_on_front;
}              vertex_obj_t;

static LIST_OF(triangle_obj_t *) triangles = NULL;
static LIST_OF(vertex_obj_t *)   vertices = NULL;
static int vertex_id = 0;
static int triangle_id = 0;
static double tol = 0;


void
crack_surf_start(void)
{
  triangle_obj_t *t;
  vertex_obj_t *v;
  
  if (triangles) {
    loop_through_list(triangles, t, triangle_obj_t *) {
      free_instance(t);
    }
    flush_list(triangles, NOT_NODES);
  }  else  {
    triangles = make_list();
  }

  if (vertices) {
    loop_through_list(vertices, v, vertex_obj_t *) {
      free_instance(v);
    }
    flush_list(vertices, NOT_NODES);
  }  else  {
    vertices = make_list();
  }

  vertex_id = triangle_id = 1;
}


static void
compute_tol(WCRec *p0, WCRec *p1);
static vertex_obj_t *
add_vertex(WCRec *pt);

void
crack_surf_add_triangle(WCRec     pts[3])
{
  triangle_obj_t *t;

  t = new_instance_of_type(triangle_obj_t);
  t->id          = triangle_id++;
  t->pts[0]      = pts[0];
  t->pts[1]      = pts[1];
  t->pts[2]      = pts[2];

  add_to_tail(triangles, t);
  tol = 0;
  compute_tol(&pts[0], &pts[1]);
  compute_tol(&pts[2], &pts[1]);
  t->vertices[0] = add_vertex(&t->pts[0])->id;
  t->vertices[1] = add_vertex(&t->pts[1])->id;
  t->vertices[2] = add_vertex(&t->pts[2])->id;
}


static int
the_same_loc(WCRec *p1, WCRec *p2, double tolerance);

static int
by_loc(WCRec *pt, vertex_obj_t *v);


static vertex_obj_t *
add_vertex(WCRec *pt)
{
  vertex_obj_t *v;

  v = (vertex_obj_t *)search_list(vertices, pt, by_loc);
  if (v == NULL) {
    v              = new_instance_of_type(vertex_obj_t);
    v->id          = vertex_id++;
    v->loc         = *pt;
    v->is_on_front = FALSE;
    add_to_tail(vertices, v);
  }
  return v;
}


static int
by_loc(WCRec *pt, vertex_obj_t *v)
{
  return the_same_loc(pt, &v->loc, tol);
}



static void
print_vertices(FILE *stream, int format);
static void
print_triangles(FILE *stream, int format);


void
crack_surf_print(FILE *stream, int format)
{
  if (triangles == NULL)
    return;

  if (format == SMF_FORMAT) fprintf (stream, " %d %d\n",
                                     list_length (vertices), list_length (triangles));


  if (format == CRACK_SURF_FORMAT)  fprintf(stream, "! vertices\n");
  print_vertices(stream, format);
  if (format == CRACK_SURF_FORMAT)  fprintf(stream, "! triangles\n");
  print_triangles(stream, format);
  if (format == CRACK_SURF_FORMAT)  fprintf(stream, "! done\n");
}




static void
compute_tol(WCRec *p0, WCRec *p1)
{
  double t =  (  fabs(p1->x - p0->x)
               + fabs(p1->y - p0->y)
               + fabs(p1->z - p0->z)) / 1000;
  tol = max(tol, t);
}



static int
the_same_loc(WCRec *p1, WCRec *p2, double tolerance)
{
  return (   fabs(p1->x - p2->x) < tolerance
          && fabs(p1->y - p2->y) < tolerance
          && fabs(p1->z - p2->z) < tolerance);
}
   


static void
print_vertices(FILE *stream, int format)
{
  vertex_obj_t *v;
  
  loop_through_list(vertices, v, vertex_obj_t *) {
    if (v->id > 0) {
      if (format == CRACK_SURF_FORMAT) {
        fprintf(stream, "V %d %g %g %g %s\n",
                v->id, v->loc.x, v->loc.y, v->loc.z,
                v->is_on_front ? "ON_FRONT" : " ");
      } else {
        fprintf(stream, "%g %g %g\n",
                v->loc.x, v->loc.y, v->loc.z);
      }
    } else {
      fprintf(stream, "Error: Dangling vertex (%g %g %g)?\n",
              v->loc.x, v->loc.y, v->loc.z);
    }
  }
}

  
static void
print_triangles(FILE *stream, int format)
{
  triangle_obj_t *t;
  
  loop_through_list(triangles, t, triangle_obj_t *) {
    if (format == CRACK_SURF_FORMAT) {
      fprintf(stream, "T %d %d %d %d\n",
              t->id, t->vertices[0], t->vertices[1], t->vertices[2]);
    } else {
      fprintf(stream, "%d %d %d\n",
              t->vertices[0]-1, t->vertices[1]-1, t->vertices[2]-1);
    }
  }
}

static int
mark_vertex(WCRec *pt);

static int
mark_vertex(WCRec *pt)
{
  add_vertex(pt)->is_on_front = TRUE;
  return TRUE;
}


void
crack_surf_add_front_segment(WCRec     pts[2])
{
  tol = 0;
  compute_tol(&pts[0], &pts[1]);
  mark_vertex(&pts[0]);
  mark_vertex(&pts[1]);
}

