
#include "hexmesh.h"



static node_t *
new_node(int id, WCRec *loc);
static hexvol_t *
new_hexvol(int id, node_t *nodes[8], GraphicObj *g_rep);
static node_t *
destroy_node(node_t *n);
static hexvol_t *
destroy_hexvol(hexvol_t *e);
static conc_load_t *
new_conc_load(int id, node_t *n, WCRec *components);
static conc_load_t *
destroy_conc_load(conc_load_t *cl);
static output_request_t *
destroy_output_request(output_request_t *or);
static output_request_t *
new_output_request(int id, node_t *node);
static void
print_fe_elem(hexmesh_model_t *model, FILE *fp, hexvol_t *e);
static void
print_efg_hexvol(hexmesh_model_t *model, FILE *fp, hexvol_t *e);
static void
print_fe_nodes(hexmesh_model_t *model, FILE *fp);
static void
print_elems(hexmesh_model_t *model, FILE *fp);
static void
print_fe_node(hexmesh_model_t *model, FILE *fp, node_t *n);
static void
print_efg_vert(hexmesh_model_t *model, FILE *fp, node_t *n);
static void
print_output_request(hexmesh_model_t *model, FILE *fp, output_request_t *or);
static void
print_conc_load(hexmesh_model_t *model, FILE *fp, conc_load_t *cl);
static int
fe_node_total(void);
static int
fe_elem_total(void);
static int
elem_total(void);
static int
have_efg_superelement(void);
static int
node_is_fixed(node_t *n);
static int
eval_curr_func(double x, double y, double z, int dir, double *fx);
static int
one_of_four(void);

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

static char *func_buf[3] = { NULL, NULL, NULL };


hexmesh_model_t *
new_hexmesh_model(char *description)
{
  hexmesh_model_t *model;
  
  model = new_instance_of_type(hexmesh_model_t);
  Assert(model != NULL);
  if (description) {
    DUPLICATE_STRING(model->description, description);
  } else {
    DUPLICATE_STRING(model->description, "unknown description");
  }
  model->thickness           = 1.0;
  model->E                   = 210000.0;
  model->nu                  = 0.3;
  model->quadrature_order    = 2;
  model->nnodes              = 0;
  model->nodes               = make_list();
  model->nhexvols            = 0;
  model->hexvols             = make_list();
  model->grep_to_hexvol_dict = dict_create_dictionary();
  model->nconc_loads         = 0;
  model->noutput_requests    = 0;
  model->conc_loads          = make_list();
  model->output_requests     = make_list();
#if USE_PIRAT
  model->node_ss             = pirat_new_pirat(100, 1.e-9);
#else
  model->node_ss             = PRaP_new_search_structure(100);
#endif  

  return model;
}


void
hexmesh_set_model_quadrature_order(hexmesh_model_t *model, int qo)
{
  model->quadrature_order = qo;
}


hexmesh_model_t *
destroy_hexmesh_model(hexmesh_model_t *model)
{
  node_t *n;
  hexvol_t *e;
  conc_load_t *cl;
  output_request_t *or;
  
  if (model->nodes != NULL && !list_empty(model->nodes)) {
    loop_through_list(model->nodes, n, node_t *) {
      destroy_node(n);
    }
    free_list(model->nodes, NOT_NODES);
  }
  if (model->hexvols != NULL && !list_empty(model->hexvols)) {
    loop_through_list(model->hexvols, e, hexvol_t *) {
      destroy_hexvol(e);
    }
    free_list(model->hexvols, NOT_NODES);
  }
  if (model->conc_loads != NULL && !list_empty(model->conc_loads)) {
    loop_through_list(model->conc_loads, cl, conc_load_t *) {
      destroy_conc_load(cl);
    }
    free_list(model->conc_loads, NOT_NODES);
  }
  if (model->output_requests != NULL && !list_empty(model->output_requests)) {
    loop_through_list(model->output_requests, or, output_request_t *) {
      destroy_output_request(or);
    }
    free_list(model->output_requests, NOT_NODES);
  }
  free_instance(model);
  return NULL;
}
  


static void
add_node_to_model(hexmesh_model_t *model, node_t *n)
{
  add_to_tail(model->nodes, n);
#if USE_PIRAT
  pirat_add_pt(model->node_ss, n, (G3D_loc_t *)&n->loc);
#else
  PRaP_add_entity(model->node_ss, n, (double *)&n->loc);
#endif
}



node_t *
hexmesh_model_add_node(hexmesh_model_t *model, WCRec *loc, double tol)
{
  node_t *n;
  
  if ((n = hexmesh_model_find_node(model, loc, tol)) == NULL) {
    n = new_node(++model->nnodes, loc);
    add_node_to_model(model, n);
  }
  return n;
}



#define EPS 0.001

#define CMP_XYZ(LOC1, LOC2)       (   (fabs(((LOC1).x)-((LOC2).x)) < EPS)  \
                                   && (fabs(((LOC1).y)-((LOC2).y)) < EPS)  \
                                   && (fabs(((LOC1).z)-((LOC2).z)) < EPS))

node_t *
hexmesh_model_find_node(hexmesh_model_t *model, WCRec *loc, double tol)
{
  node_t *n, *then;
  int status;
#if USE_PIRAT
  G3D_box_t range;
#else
  PRaP_range_t range;
#endif
  double cd, dist;
  WCRec d;
#if USE_PIRAT
#define SUPRNG(at, drange) \
  {G3D_INIT_BOX(range, at); G3D_INFLATE_BOX(range, drange);}
#else
#define SUPRNG(at, drange)           \
  {                                  \
     range.lower[0] = at.x - drange; \
     range.lower[1] = at.y - drange; \
     range.lower[2] = at.z - drange; \
     range.upper[0] = at.x + drange; \
     range.upper[1] = at.y + drange; \
     range.upper[2] = at.z + drange; \
  }  
#endif

  SUPRNG((*loc), tol);
#if USE_PIRAT
  status = pirat_search_range(model->node_ss, &range, INT_MAX);
#else
  status = PRaP_search_range(model->node_ss, &range);
#endif
  if (!status)
    return NULL;

  then = NULL;
  dist = FLT_MAX;
#if USE_PIRAT
  n = pirat_first_in_range(model->node_ss);
#else
  n = PRaP_first_entity_in_range(model->node_ss);
#endif
  while (n != NULL) {
    VECMAC_3_VEC(d, =, (*n).loc, -, (*loc));
    cd = VECMAC_NORM(d);
    if (cd < dist) {
      then = n;
      dist = cd;
    }
#if USE_PIRAT
    n = pirat_next_in_range(model->node_ss);
#else
    n = (node_t *)PRaP_next_entity_in_range(model->node_ss);
#endif
  }

  return then;
}


hexvol_t *
hexmesh_model_add_hexvol(hexmesh_model_t *model, node_t *nodes[8],
                       GraphicObj *g_rep)
{
  hexvol_t *e;

  Assert(nodes[0]);
  Assert(nodes[1]);
  Assert(nodes[2]);
  Assert(nodes[3]);
  Assert(nodes[4]);
  Assert(nodes[5]);
  Assert(nodes[6]);
  Assert(nodes[7]);
  e = new_hexvol(++model->nhexvols, nodes, g_rep);
  add_to_tail(model->hexvols, e);
  dict_add_obj_handle_name(model->grep_to_hexvol_dict, e, g_rep);
  
  return e;
}


void
hexmesh_set_hexvol_partition(hexmesh_model_t *model, hexvol_t *h, int partition)
{
  h->partition = partition;
}


void
hexmesh_set_hexvol_mat_id (hexmesh_model_t *model, hexvol_t *h, int matid)
{
  h->matid = matid;
}


int
hexmesh_get_hexvol_partition(hexmesh_model_t *model, hexvol_t *h)
{
  return h->partition;
}


void
hexmesh_set_hexvol_quadrature_order(hexmesh_model_t *model, hexvol_t *h,
                                    int quadrature_order)
{
  h->quadrature_order = quadrature_order;
}



void
hexmesh_set_hexvol_subsplit(hexmesh_model_t *model, hexvol_t *h,
                                    int subsplit)
{
  h->split = subsplit;
}

void
hexmesh_model_change_to_efg(hexmesh_model_t *model, hexvol_t *el)
{
  Assert(el);
  el->desired_type = el->type = EFG_HEXVOL; 
}


void 
hexmesh_model_change_to_fe(hexmesh_model_t *model, hexvol_t *el)
{
  Assert(el);
  el->desired_type = el->type = FE_HEXVOL;
}



static void
set_prescr_displ(node_t *n, int dir);

void
hexmesh_model_add_constraint(hexmesh_model_t *model, int type, LIST nodes)
{
  node_t *n;

  loop_through_list(nodes, n, node_t *) {
    switch (type) {
    case 'x':
    case 'X':
      n->constraint[0] = CONSTRAINT_FIXED;
      set_prescr_displ(n, 0);
      break;
    case 'y':
    case 'Y':
      n->constraint[1] = CONSTRAINT_FIXED;
      set_prescr_displ(n, 1);
      break;
    case 'z':
    case 'Z':
      n->constraint[2] = CONSTRAINT_FIXED;
      set_prescr_displ(n, 2);
      break;
    case 'a':
    case 'A':
      n->constraint[0] = n->constraint[1] = n->constraint[2]
        = CONSTRAINT_FIXED;
      set_prescr_displ(n, 0);
      set_prescr_displ(n, 1);
      set_prescr_displ(n, 2);
      break;
    case 'f':
    case 'F':
      n->constraint[0] = n->constraint[1] = n->constraint[2] = CONSTRAINT_FREE;
      n->prescr_displ[0] = n->prescr_displ[1] = n->prescr_displ[2] = 0;
      break;
    default:
      break;                    /* VOID */
    }
  }      
}


static void
set_prescr_displ(node_t *n, int dir)
{
  DUPLICATE_STRING(n->symbolic_prescr_displ[dir], func_buf[dir]);
  eval_curr_func(n->loc.x, n->loc.y, n->loc.z, dir, &n->prescr_displ[dir]);
}


static int
node_is_fixed(node_t *n)
{ 
  return (   (n->constraint[0] == CONSTRAINT_FIXED)
          || (n->constraint[1] == CONSTRAINT_FIXED)
          || (n->constraint[2] == CONSTRAINT_FIXED));
}



/* The faces are: */
/* face #      vertices (counterclockwise when seen from outside) */
static int face_vertex_indices[6 /* face # */][4 /* vertex index */] = {
/* 0 */ { 0, 3, 2, 1 },
/* 1 */ { 0, 1, 5, 4 },
/* 2 */ { 1, 2, 6, 5 },
/* 3 */ { 2, 3, 7, 6 },
/* 4 */ { 3, 0, 4, 7 },
/* 5 */ { 4, 5, 6, 7 }
      };

static surface_traction_t *
new_surf_traction(node_t *n0, node_t *n1, node_t *n2, node_t *n3);


static void
surf_tractions_on_hex (hexvol_t *e,
                       hexmesh_model_t *model,
                       int surf_traction_tag)
{
  int f;

  for (f = 0; f < 6; f++) {
    if (   (e->nodes[face_vertex_indices[f][0]]->surf_traction_tag
            == surf_traction_tag)
        && (e->nodes[face_vertex_indices[f][1]]->surf_traction_tag
            == surf_traction_tag)
        && (e->nodes[face_vertex_indices[f][2]]->surf_traction_tag
            == surf_traction_tag)
        && (e->nodes[face_vertex_indices[f][3]]->surf_traction_tag
            == surf_traction_tag)
        ) {
      e->face_traction[f]
        = new_surf_traction(e->nodes[face_vertex_indices[f][0]],
                            e->nodes[face_vertex_indices[f][1]],
                            e->nodes[face_vertex_indices[f][2]],
                            e->nodes[face_vertex_indices[f][3]]);
    }
  }
}



void
hexmesh_model_add_surf_tractions(hexmesh_model_t *model,
                               LIST_OF(node_t *) nodes)
{
  hexvol_t *e;
  node_t *n;
  static int surf_traction_tag = 0;

  surf_traction_tag++;
  loop_through_list(nodes, n, node_t *) {
    n->surf_traction_tag = surf_traction_tag;
  }
  
  loop_through_list(model->hexvols, e, hexvol_t *) {
    surf_tractions_on_hex (e, model, surf_traction_tag);
  }
}





int
hexmesh_set_function(int dir, char *expr)
{
  double f;
  
  if (expr != NULL) {
    switch (dir) {
    case 'x':
      func_buf[0] = realloc_array(func_buf[0], strlen(expr) + 1, char);
      Assert(func_buf[0]);
      strcpy(func_buf[0], expr);
      return eval_curr_func(0, 0, 0, 0, &f);
      break;
    case 'y':
      func_buf[1] = realloc_array(func_buf[1], strlen(expr) + 1, char);
      Assert(func_buf[1]);
      strcpy(func_buf[1], expr);
      return eval_curr_func(0, 0, 0, 1, &f);
      break;
    case 'z':
      func_buf[2] = realloc_array(func_buf[2], strlen(expr) + 1, char);
      Assert(func_buf[2]);
      strcpy(func_buf[2], expr);
      return eval_curr_func(0, 0, 0, 2, &f);
      break;
    }    
  }
  return EXPEVAL_E_OK;
}


static int
eval_curr_func(double x, double y, double z, int dir, double *f)
{
  int stat = EXPEVAL_E_OK;
  
  expeval_set_value("x", x);
  expeval_set_value("y", y);
  expeval_set_value("z", z);

  if (dir >= 0 && dir < 3) {
    if (func_buf[dir] == NULL) {
      *f = 0;
    } else {
      stat = expeval_eval(func_buf[dir], f);
      if (stat != EXPEVAL_E_OK) {
        *f = 0;  
      }
    }
  } else {
    fprintf(stderr, "Bad input %c in eval_curr_func()\n", dir);
    stat = EXPEVAL_E_OK;
  }
  return stat;
}



static surface_traction_t *
new_surf_traction(node_t *n0, node_t *n1, node_t *n2, node_t *n3)
{
  surface_traction_t *st;

  st = new_instance_of_type(surface_traction_t);
  Assert(st != NULL);

  DUPLICATE_STRING(st->symbolic_traction[0], func_buf[0]);
  DUPLICATE_STRING(st->symbolic_traction[1], func_buf[1]);
  DUPLICATE_STRING(st->symbolic_traction[2], func_buf[2]);

  eval_curr_func(n0->loc.x, n0->loc.y, n0->loc.z, 0,
                 &st->node_traction[0].x);
  eval_curr_func(n0->loc.x, n0->loc.y, n0->loc.z, 1,
                 &st->node_traction[0].y);
  eval_curr_func(n0->loc.x, n0->loc.y, n0->loc.z, 2,
                 &st->node_traction[0].z);
  
  eval_curr_func(n1->loc.x, n1->loc.y, n1->loc.z, 0,
                 &st->node_traction[1].x);
  eval_curr_func(n1->loc.x, n1->loc.y, n1->loc.z, 1,
                 &st->node_traction[1].y);
  eval_curr_func(n1->loc.x, n1->loc.y, n1->loc.z, 2,
                 &st->node_traction[1].z);
  
  eval_curr_func(n2->loc.x, n2->loc.y, n2->loc.z, 0,
                 &st->node_traction[2].x);
  eval_curr_func(n2->loc.x, n2->loc.y, n2->loc.z, 1,
                 &st->node_traction[2].y);
  eval_curr_func(n2->loc.x, n2->loc.y, n2->loc.z, 2,
                 &st->node_traction[2].z);

  eval_curr_func(n3->loc.x, n3->loc.y, n3->loc.z, 0,
                 &st->node_traction[3].x);
  eval_curr_func(n3->loc.x, n3->loc.y, n3->loc.z, 1,
                 &st->node_traction[3].y);
  eval_curr_func(n3->loc.x, n3->loc.y, n3->loc.z, 2,
                 &st->node_traction[3].z);
  
  return st;
}


void
hexmesh_model_add_conc_load(hexmesh_model_t *model, node_t *node,
                          WCRec *components)
{
  conc_load_t *cl;

  cl = new_conc_load(++model->nconc_loads, node, components);
  Assert(cl);
  add_to_tail(model->conc_loads, cl);
}




static node_t *
new_node(int id, WCRec *loc)
{
  node_t *n;

  n = new_instance_of_type(node_t);
  Assert(n);
  n->fe_id                  = id;
  n->efg_id                 = -1;
  n->type                   = FE_NODE;
  n->incident_to_fe_hexvol  = TRUE;
  n->incident_to_efg_hexvol = FALSE;
  n->loc                    = *loc;
  n->constraint[0] = n->constraint[1] = n->constraint[2] = CONSTRAINT_FREE;
  n->prescr_displ[0] = n->prescr_displ[1] = n->prescr_displ[2] = 0;
  n->g_rep = CreateMarker3D(&n->loc);
  EGWithMaskChangeAttributes(  LAYER_MASK
                             | COLOR_MASK
                             | MSIZE_MASK
                             | MTYPE_MASK, n->g_rep);
  EGSetPickableMask(n->g_rep, NODE_PICK_MASK);
  EMNoDrawAddGraphicsToModel(ESIModel(), n->g_rep);
  n->iface_id = -1;
  n->d_m = 0;
  
  return n;
}



static node_t *
destroy_node(node_t *n)
{
  EGDeepDestroyGraphics(n->g_rep);
  free_instance(n);
  return NULL;
}



static void
reorder_nodes(hexvol_t *e);

static hexvol_t *
new_hexvol(int id, node_t *nodes[8], GraphicObj *g_rep)
{
  hexvol_t *e;

  e = new_instance_of_type(hexvol_t);
  Assert(e);
  e->fe_id    = id;
  e->efg_id   = -1;
  e->desired_type = e->type = FE_HEXVOL;
  e->nodes[0] = nodes[0];
  e->nodes[1] = nodes[1];
  e->nodes[2] = nodes[2];
  e->nodes[3] = nodes[3];
  e->nodes[4] = nodes[4];
  e->nodes[5] = nodes[5];
  e->nodes[6] = nodes[6];
  e->nodes[7] = nodes[7];
  reorder_nodes(e);
  e->g_rep = g_rep;
  EGSetPickableMask(e->g_rep, HEX_PICK_MASK);
  e->partition = -1;
  e->is_ghost  = FALSE;
  {
    int i;
    for (i = 0; i < 6; i++)  e->face_traction[i] = NULL;
  }
  e->d_m = 0;
  e->quadrature_order = 0;
  e->split   = 0;
  e->matid   = 1;

  return e;
}


static void
reorder_nodes(hexvol_t *e)
{
  CkitVector3 v01half, v12half, v23half, v30half, v1, v2, v3, v;
  CkitVector3 p[8];
  int i;
#define SWP(n1, n2) { node_t *t; t = (n1); (n1) = (n2); (n2) = t; }
#define SWAP_NODES()                            \
  {                                             \
    SWP(e->nodes[0], e->nodes[4]);              \
    SWP(e->nodes[1], e->nodes[5]);              \
    SWP(e->nodes[2], e->nodes[6]);              \
    SWP(e->nodes[3], e->nodes[7]);              \
  }
    
  FOR(i, 8) { p[i].x = e->nodes[i]->loc.x;
              p[i].y = e->nodes[i]->loc.y;
              p[i].z = e->nodes[i]->loc.z;
            }
  v3_lerp(&p[0], &p[1], 0.5, &v01half);
  v3_lerp(&p[1], &p[2], 0.5, &v12half);
  v3_lerp(&p[2], &p[3], 0.5, &v23half);
  v3_lerp(&p[3], &p[0], 0.5, &v30half);
  v3_sub(&v12half, &v30half, &v1);
  v3_sub(&v23half, &v01half, &v2);
  v3_cross(&v1, &v2, &v3);

  v3_sub(&p[4], &p[0], &v);
  if (v3_dot(&v3, &v) < 0) { SWAP_NODES(); return; }
  v3_sub(&p[5], &p[1], &v);
  if (v3_dot(&v3, &v) < 0) { SWAP_NODES(); return; }
  v3_sub(&p[6], &p[2], &v);
  if (v3_dot(&v3, &v) < 0) { SWAP_NODES(); return; }
  v3_sub(&p[7], &p[3], &v);
  if (v3_dot(&v3, &v) < 0) { SWAP_NODES(); return; }
}


static hexvol_t *
destroy_hexvol(hexvol_t *e)
{
  free_instance(e);
  return NULL;
}





static conc_load_t *
new_conc_load(int id, node_t *node, WCRec *components)
{
  conc_load_t *cl;

  cl = new_instance_of_type(conc_load_t);
  Assert(cl);
  cl->id                 = id;
  cl->node               = node;
  cl->components         = *components;
  cl->included_in_output = FALSE;
  return cl;
}



static conc_load_t *
destroy_conc_load(conc_load_t *cl)
{
  free_instance(cl);
  return NULL;
}

static output_request_t *
destroy_output_request(output_request_t *or)
{
  free_instance(or);
  return NULL;
}


static output_request_t *
new_output_request(int id, node_t *node)
{
  output_request_t *or;

  or = new_instance_of_type(output_request_t);
  Assert(or);
  or->id         = id;
  or->node       = node;

  return or;
}



#if 0
static conc_load_t *
have_conc_load_at_node(hexmesh_model_t *model, node_t *n)
{
  conc_load_t *cl;

  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    if (cl->node == n)
      return cl;
  }
  return NULL;
}
#endif


static int
hexmesh_nloads(hexmesh_model_t *model);



static void
build_rep(hexmesh_model_t *model);
static void
print_fe_distrib_load(hexmesh_model_t *model, FILE *fp, hexvol_t *e);


void
hexmesh_model_print(hexmesh_model_t *model, FILE *fp)
{
  conc_load_t *cl;
  output_request_t *or;

  build_rep(model);
  
  if (model->description != NULL)
    fprintf(fp, "%s\n", model->description);
  else
    fprintf(fp, " model description unknown\n");
  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "!     INPUT FILE FOR WASP; GENERATED BY HEXMESH \n");
  fprintf(fp, "!                (C) 1995 Petr Krysl \n");
  fprintf(fp, "! \n");

  fprintf(fp, "! #nodes #elements #materials #loads #time_funcs  \n");
  fprintf(fp, "   %d      %d        %d        %d        %d\n",
          fe_node_total(), elem_total(), 1, hexmesh_nloads(model), 1);
  

  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "OPTIONS \n");
  fprintf(fp, "!  option_tag (O, T, S)     value \n");
  fprintf(fp, "!  END OPTIONS \n");
  fprintf(fp, "&\n");


  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "MATERIAL \n");
  fprintf(fp, "   5  \n");
  fprintf(fp, "!  id   E        nu    rho\n");
  fprintf(fp, "   1   %g   %g   %g \n", model->E, model->nu, model->rho);
  fprintf(fp, "!  END MATERIAL \n");
  fprintf(fp, "&\n");


  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "TIME_FUNCTIONS \n");
  fprintf(fp, "  2  ! tabular \n");
  fprintf(fp, "! func#   #pairs \n");
  fprintf(fp, "   1        3\n");
  fprintf(fp, "! time     val \n");
  fprintf(fp, "! Heaviside step function:\n");
  fprintf(fp, "   0        1\n");
  fprintf(fp, "   1        1\n");
  fprintf(fp, "   10000000 1\n");
  fprintf(fp, "!  END TIME_FUNCTIONS \n");
  fprintf(fp, "&\n");


  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "NODES\n");

  fprintf(fp, "! id       x       y       z  CONSTRAIN: X  Y  Z (0=fix)\n");
  print_fe_nodes(model, fp);
  fprintf(fp, "!  END NODES \n");
  fprintf(fp, "&\n");

  print_elems(model, fp);

  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "LOADS \n");
  fprintf(fp, "! load type (1 = nodal) \n");
  fprintf(fp, "! node#  dir (0-5)  amplitude  #time_func \n");
  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    print_conc_load(model, fp, cl);
  }
  {
    hexvol_t *h;
    loop_through_list(model->hexvols, h, hexvol_t *) {
      if (HV_IS_FE(h))
        print_fe_distrib_load(model, fp, h);
    }
  }
  fprintf(fp, "!  END LOADS \n");
  fprintf(fp, "&\n");

  
  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "! OUTPUT REQUESTS \n");
  fprintf(fp, "! vertex# \n");
  if (!list_empty(model->output_requests)) {
    loop_through_list(model->output_requests, or, output_request_t *) {
      print_output_request(model, fp, or);
    }
  }
  fprintf(fp, "!  END OUTPUT REQUESTS \n");


  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "HISTORY \n");
  fprintf(fp, "!  client_tag (N, E, W)     entity# (0 for energy (W)) \n");
  fprintf(fp, "!  END HISTORY \n");
  fprintf(fp, "&\n");
  

  fprintf(fp, "!  END DATA \n");
}




void
hexmesh_model_set_properties(hexmesh_model_t *model,
                        double E, double nu, double rho)
{
  model->E         = E;
  model->nu        = nu;
  model->rho       = rho;
}


void
hexmesh_model_add_output_request(hexmesh_model_t *model, node_t *n)
{
  output_request_t *or;

  or = new_output_request(++model->noutput_requests, n);
  add_to_tail(model->output_requests, or);
}


static int
have_node(hexvol_t *e, node_t *n);

hexvol_t *
hexmesh_model_find_hexvol(hexmesh_model_t *model, node_t *nodes[8])
{
  hexvol_t *e;
  
  loop_through_list(model->hexvols, e, hexvol_t *) {
    if (   have_node(e, nodes[0])
        && have_node(e, nodes[1])
        && have_node(e, nodes[2])
        && have_node(e, nodes[3])
        && have_node(e, nodes[4])
        && have_node(e, nodes[5])
        && have_node(e, nodes[6])
        && have_node(e, nodes[7]))
      return e;
  }
  return NULL;
}



hexvol_t *
hexmesh_model_hexvol_by_g_rep(hexmesh_model_t *model, GraphicObj *g_rep)
{
  hexvol_t *e;

  e = (hexvol_t *)dict_find_obj_with_handle_name(model->grep_to_hexvol_dict,
                                                 g_rep);
  return e;
}


static int
have_node(hexvol_t *e, node_t *n)
{
  return (   (e->nodes[0] == n)
          || (e->nodes[1] == n)
          || (e->nodes[2] == n)
          || (e->nodes[3] == n)
          || (e->nodes[4] == n)
          || (e->nodes[5] == n)
          || (e->nodes[6] == n)
          || (e->nodes[7] == n));
}



static void
print_fe_node(hexmesh_model_t *model, FILE *fp, node_t *n)
{
  fprintf(fp, "  %d   %g   %g  %g       %d   %d   %d\n",
          n->fe_id, n->loc.x, n->loc.y, n->loc.z, 
          n->constraint[0],
          n->constraint[1],
          n->constraint[2]);
}



static void
print_efg_vert(hexmesh_model_t *model, FILE *fp, node_t *n)
{
  fprintf(fp, "      %d   %g   %g  %g\n",
          n->efg_id, n->loc.x, n->loc.y, n->loc.z);
}



static void
print_conc_load(hexmesh_model_t *model, FILE *fp, conc_load_t *cl)
{
  if (cl->components.x != 0) {
    fprintf(fp, "   1 \n");
    fprintf(fp, "  %d   DX   %g   %d\n",
            cl->node->fe_id, cl->components.x, 1);
  }
  if (cl->components.y != 0) {
    fprintf(fp, "   1 \n");
    fprintf(fp, "  %d   DY   %g   %d\n",
            cl->node->fe_id, cl->components.y, 1);
  }
  if (cl->components.z != 0) {
    fprintf(fp, "   1 \n");
    fprintf(fp, "  %d   DZ   %g   %d\n",
            cl->node->fe_id, cl->components.z, 1);
  }
}


static int
hexmesh_nloads(hexmesh_model_t *model)
{
  conc_load_t *cl;
  int nloads = 0;

  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    if (cl->components.x != 0) nloads++;
    if (cl->components.y != 0) nloads++;
    if (cl->components.z != 0) nloads++;
  }
  return nloads;
}


static void
print_fe_elem(hexmesh_model_t *model, FILE *fp, hexvol_t *e)
{
  fprintf(fp, "%d %d %d %d %d %d %d %d %d %d\n",
          e->fe_id,
          e->nodes[0]->fe_id,
          e->nodes[1]->fe_id,
          e->nodes[2]->fe_id,
          e->nodes[3]->fe_id,
          e->nodes[4]->fe_id,
          e->nodes[5]->fe_id,
          e->nodes[6]->fe_id,
          e->nodes[7]->fe_id,
          e->matid);
}


static void
print_output_request(hexmesh_model_t *model, FILE *fp, output_request_t *or)
{
  fprintf(fp, "!   %d\n", or->node->fe_id);
}



static void
adjust_types(hexmesh_model_t *model);

static void
adjust_types(hexmesh_model_t *model)
{
  node_t *n;
  hexvol_t *h;
  int i;
#define PURE_FE(h, I)  (N_INC_PURE_FE(h->nodes[I]))
#define MIXED(h, I)    (N_INC_MIXED(h->nodes[I]))
#define PURE_EFG(h, I) (N_INC_PURE_EFG(h->nodes[I]))
#define HEXVOL_TYPE(h)                                  \
  {                                                     \
    if        (   (PURE_FE(h, 0) || MIXED(h, 0))        \
               && (PURE_FE(h, 1) || MIXED(h, 1))        \
               && (PURE_FE(h, 2) || MIXED(h, 2))        \
               && (PURE_FE(h, 3) || MIXED(h, 3))        \
               && (PURE_FE(h, 4) || MIXED(h, 4))        \
               && (PURE_FE(h, 5) || MIXED(h, 5))        \
               && (PURE_FE(h, 6) || MIXED(h, 6))        \
               && (PURE_FE(h, 7) || MIXED(h, 7))        \
               ) {                                      \
      h->type = FE_HEXVOL;                              \
    } else if (   PURE_EFG(h, 0)                        \
               && PURE_EFG(h, 1)                        \
               && PURE_EFG(h, 2)                        \
               && PURE_EFG(h, 3)                        \
               && PURE_EFG(h, 4)                        \
               && PURE_EFG(h, 5)                        \
               && PURE_EFG(h, 6)                        \
               && PURE_EFG(h, 7)                        \
               ) {                                      \
      h->type = EFG_HEXVOL;                             \
    } else {                                            \
      h->type = EFG_IFACE_HEXVOL;                       \
    }                                                   \
  }
#define INIT_NODE_PROPS()                               \
  {                                                     \
    loop_through_list(model->nodes, n, node_t *) {      \
      n->type = FE_NODE;                                \
      n->incident_to_fe_hexvol = FALSE;                 \
      n->incident_to_efg_hexvol = FALSE;                \
      n->incident_to_iface_hexvol = FALSE;              \
    }                                                   \
  }
#define INIT_HEXVOL_PROPS()                             \
  {                                                     \
    loop_through_list(model->hexvols, h, hexvol_t *) {  \
      h->type = h->desired_type;                        \
    }                                                   \
  }
#define SET_EFG_INC(n) { (n)->incident_to_efg_hexvol = TRUE; }
#define SET_FE_INC(n)  { (n)->incident_to_fe_hexvol = TRUE; }
#define SET_IFACE_INC(n)  { (n)->incident_to_iface_hexvol = TRUE; }

  INIT_NODE_PROPS();
  INIT_HEXVOL_PROPS();
  loop_through_list(model->hexvols, h, hexvol_t *) {
    if        (HV_IS_EFG(h)) {
      FOR(i, 8) {
        SET_EFG_INC(h->nodes[i]);
        if (node_is_fixed(h->nodes[i]))
          { SET_FE_INC(h->nodes[i]); }
      }
    } else if (HV_IS_FE(h)) {
      FOR(i, 8) { SET_FE_INC(h->nodes[i]); }
    }
  }
  loop_through_list(model->hexvols, h, hexvol_t *) {
    HEXVOL_TYPE(h);
    if        (HV_IS_IFACE_EFG(h)) {
      FOR(i, 8) {
        SET_IFACE_INC(h->nodes[i]);
        if        (MIXED(h, i)) {
          h->nodes[i]->type = FE_EFG_NODE;
        } else if (PURE_EFG(h, i)) {
          h->nodes[i]->type = EFG_NODE;
        }
      }
    } else if (HV_IS_PURE_EFG(h)) {
      FOR(i, 8) { h->nodes[i]->type = EFG_NODE; }
    }
  }
}



static void
print_fe_nodes(hexmesh_model_t *model, FILE *fp)
{
  node_t *n;
  
  loop_through_list(model->nodes, n, node_t *) {
    if (NODE_IS_FE(n)) 
      print_fe_node(model, fp, n);
  }
}

static void
print_efg_superelem(hexmesh_model_t *model, FILE *fp);


static void
print_elems(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *h;

  if (fe_elem_total() > 0) {
    fprintf(fp, "!  ======================================================\n");
    fprintf(fp, "ELEM\n");
    fprintf(fp, "   6  ! 2pt and 8pt isoparametric hexahedrons\n");
    fprintf(fp, "!  id    n1 n2 n3 n4     n5 n6 n7 n8    material \n");
    fprintf(fp, "USE_2PT_RULE ! enable 2pt quadrature by default\n");
    loop_through_list(model->hexvols, h, hexvol_t *) {
      if (HV_IS_FE(h))
        print_fe_elem(model, fp, h);
    }
    fprintf(fp, "!  END ELEM \n");
    fprintf(fp, "&\n");
  }
  if (have_efg_superelement()) {
    fprintf(fp, "!  ======================================================\n");
    fprintf(fp, "ELEM\n");
    print_efg_superelem(model, fp);
    fprintf(fp, "!  END ELEM \n");
    fprintf(fp, "&\n");
  }
}


static void
print_efg_distrib_load(hexmesh_model_t *model, FILE *fp, hexvol_t *e);


static void
print_efg_superelem(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *h;
  node_t *n;
  
  fprintf(fp, "    5  ! EFG superelement\n");
  fprintf(fp, "EFGSUPERELEM %d\n", fe_elem_total()+1);
  fprintf(fp, "!  VSS_BUCKET ?\n");
  fprintf(fp, "!  EFGN_BUCKET ?\n");
  fprintf(fp, "!  TOLERANCE ?\n");
  fprintf(fp, "!  CONVEX\n");
  fprintf(fp, "!  NON_CONVEX\n");
  fprintf(fp, "!  CONVEX_WITH_CRACKS\n");
  fprintf(fp, "!  GENERATE_NODES_INSIDE_ONLY\n");
  fprintf(fp, "!  GENERATE_NODES_IN_VOLUME lx ly lz ux uy uz\n");
  fprintf(fp, "!  SPACING       ?\n");
  fprintf(fp, "  NODES_PER_SIDE ?\n");
  fprintf(fp, "  SUPPORT_SIZE_MULTIPLIER ?\n");
  fprintf(fp, "  COMPUTE_SUPPORT_SIZE\n");
  fprintf(fp, "  SUGGESTED_NEIGHBORS 8\n");
  fprintf(fp, "  WEIGHT_FUNC_TYPE SMOOTH_QUADRATIC_SPHERICAL\n");
  fprintf(fp, "  QUADRATURE    %d\n", model->quadrature_order);
  fprintf(fp, "!  ENABLE_2PT_QUADRATURE\n");
  fprintf(fp, "  MATERIAL \n");
  fprintf(fp, "  3 ! Instantaneous final rotation (J2 elastoplastic)\n");
  fprintf(fp, "  1 E ? NU ? RHO ? SIGMA_Y ? HARDENING_PARAM ? DELTA ?\n");
  fprintf(fp, "!  Taylor bar material:\n");
  fprintf(fp, "!     E 117000 NU 0.35 RHO 8.93e-9\n");
  fprintf(fp, "!     SIGMA_Y 400 HARDENING_PARAM 100 DELTA 1\n");
  fprintf(fp, "  VERT  ! vert-num x y z\n");
  loop_through_list(model->nodes, n, node_t *) {
    if (NODE_IS_EFG(n)) {
      print_efg_vert(model, fp, n);
    }
  }
  fprintf(fp, "  END VERT \n");
  fprintf(fp, "  CONNECT  ! vertex-num   fe-node-num \n");
  loop_through_list(model->nodes, n, node_t *) {
    if (N_INC_MIXED(n)) {
      fprintf(fp, "     %d   %d\n", n->efg_id, n->fe_id);
    }
  }
  fprintf(fp, "  END CONNECT \n");
  fprintf(fp, "  HEXVOL ! id  n1 n2 n3 n4  n5 n6 n7 n8 [SPLIT [SIZE] num]\n");
  loop_through_list(model->hexvols, h, hexvol_t *) {
    if (HV_IS_EFG(h))
      print_efg_hexvol(model, fp, h);
  }
  fprintf(fp, "  END HEXVOL \n");
  fprintf(fp, "   DISTRIBL\n");
  loop_through_list(model->hexvols, h, hexvol_t *) {
    if (HV_IS_EFG(h))
      print_efg_distrib_load(model, fp, h);
  }
  fprintf(fp, "   END DISTRIBL\n");
  fprintf(fp, "!  CONCL\n");
  fprintf(fp, "!  END CONCL\n");
  fprintf(fp, "!  CONTACT\n");
  fprintf(fp, "!  CRACK\n");
  fprintf(fp, "!    TRIANGULATED\n");
  fprintf(fp, "!       ID 1\n");
  fprintf(fp, "!    END TRIANGULATED\n");
  fprintf(fp, "!  END CRACK\n");
  fprintf(fp, "!  END CONTACTL\n");
  fprintf(fp, "END\n");
}


static void
print_efg_hexvol(hexmesh_model_t *model, FILE *fp, hexvol_t *e)
{
  fprintf(fp, "     %d    %d %d %d %d     %d %d %d %d ",
          e->efg_id,
          e->nodes[0]->efg_id,
          e->nodes[1]->efg_id,
          e->nodes[2]->efg_id,
          e->nodes[3]->efg_id,
          e->nodes[4]->efg_id,
          e->nodes[5]->efg_id,
          e->nodes[6]->efg_id,
          e->nodes[7]->efg_id);
  if (e->split > 1)
    fprintf(fp, " SPLIT %d", e->split);
  if (e->quadrature_order > 0)
    fprintf(fp, " QO %d", e->quadrature_order);
  fprintf(fp, "\n");
}


static void
print_efg_distrib_load(hexmesh_model_t *model, FILE *fp, hexvol_t *e)
{
  int f, i;
  
  FOR(f, 6) {
    if (e->face_traction[f] != NULL) {
        fprintf(fp, "1 DEAD_PRESSURE 1 qr 2\n");
        FOR (i, 4) {
          fprintf(fp, "%g %g %g\n",
                  e->nodes[face_vertex_indices[f][i]]->loc.x,
                  e->nodes[face_vertex_indices[f][i]]->loc.y,
                  e->nodes[face_vertex_indices[f][i]]->loc.z);
      }
    }
  }
}

static double
face_area(hexvol_t *e, int f);


static void
print_fe_distrib_load(hexmesh_model_t *model, FILE *fp, hexvol_t *e)
{
  int f, i;
  double area, force;
  /* h->face_traction[f]->node_traction[n].x */
  
  FOR(f, 6) {
    if (e->face_traction[f] != NULL) {
      area = face_area(e, f);
      force = (  e->face_traction[f]->node_traction[0].x
               + e->face_traction[f]->node_traction[1].x
               + e->face_traction[f]->node_traction[2].x
               + e->face_traction[f]->node_traction[3].x) / 4 * area / 4;
      if (force != 0) {
        FOR (i, 4) {
          fprintf(fp, "1\n");
          fprintf(fp, "%d DX %g 1\n",
                  e->nodes[face_vertex_indices[f][i]]->fe_id,
                  force);
        }
      }
      force = (  e->face_traction[f]->node_traction[0].y
               + e->face_traction[f]->node_traction[1].y
               + e->face_traction[f]->node_traction[2].y
               + e->face_traction[f]->node_traction[3].y) / 4 * area / 4;
      if (force != 0) {
        FOR (i, 4) {
          fprintf(fp, "1\n");
          fprintf(fp, "%d DY %g 1\n",
                  e->nodes[face_vertex_indices[f][i]]->fe_id,
                  force);
        }
      }
      force = (  e->face_traction[f]->node_traction[0].z
               + e->face_traction[f]->node_traction[1].z
               + e->face_traction[f]->node_traction[2].z
               + e->face_traction[f]->node_traction[3].z) / 4 * area / 4;
      if (force != 0) {
        FOR (i, 4) {
          fprintf(fp, "1\n");
          fprintf(fp, "%d DZ %g 1\n",
                  e->nodes[face_vertex_indices[f][i]]->fe_id,
                  force);
        }
      }
    }
  }
}


static double
face_area(hexvol_t *e, int f)
{
  G3D_vect_t AB, AC, ABAC;
  double area;
  
  VECMAC_3_VEC(AB, =,
               e->nodes[face_vertex_indices[f][1]]->loc, -,
               e->nodes[face_vertex_indices[f][0]]->loc);
  VECMAC_3_VEC(AC, =,
               e->nodes[face_vertex_indices[f][2]]->loc, -,
               e->nodes[face_vertex_indices[f][0]]->loc);
  VECMAC_SCALAR(ABAC, =, 0);
  VECMAC_CROSS(ABAC, =, AB, x, AC);
  area = VECMAC_NORM(ABAC);
  
  VECMAC_3_VEC(AB, =,
               e->nodes[face_vertex_indices[f][2]]->loc, -,
               e->nodes[face_vertex_indices[f][0]]->loc);
  VECMAC_3_VEC(AC, =,
               e->nodes[face_vertex_indices[f][3]]->loc, -,
               e->nodes[face_vertex_indices[f][0]]->loc);
  VECMAC_SCALAR(ABAC, =, 0);
  VECMAC_CROSS(ABAC, =, AB, x, AC);
  area += VECMAC_NORM(ABAC);

  return area / 2;
}





static void
renumber_nodes(hexmesh_model_t *model);
static void
renumber_elems(hexmesh_model_t *model);

static int fenode_total;
static int efgnode_total;
static int feelem_total;
static int efgvol_total;

static void
build_rep(hexmesh_model_t *model)
{
  adjust_types(model);
  renumber_nodes(model);
  renumber_elems(model);
}

static int
fe_node_total(void)
{
  return fenode_total;
}

static int
fe_elem_total(void)
{
  return feelem_total;
}


static int
elem_total(void)
{
  return feelem_total + (have_efg_superelement()? 1: 0);
}


static int
have_efg_superelement(void)
{
  return efgvol_total > 0;
}

static void
renumber_nodes(hexmesh_model_t *model)
{
  node_t *n;
  int ntot;

  fenode_total = 0;
  efgnode_total = 0;
  ntot = list_length(model->nodes);
  loop_through_list(model->nodes, n, node_t *) {
    if (NODE_IS_FE(n))
      n->fe_id = ++fenode_total;
    if (NODE_IS_EFG(n))
      n->efg_id = ++efgnode_total;
  }
}


static void
renumber_elems(hexmesh_model_t *model)
{
  hexvol_t *h;
  int hvtot;

  feelem_total = 0;
  efgvol_total = 0;
  hvtot = list_length(model->hexvols);
  loop_through_list(model->hexvols, h, hexvol_t *) {
    if (HV_IS_FE(h)) {
      h->fe_id = ++feelem_total;
      h->efg_id = -1;
    }
    if (HV_IS_EFG(h)) {
      h->fe_id = -1;
      h->efg_id = ++efgvol_total;
    }
  }
  Assert(hvtot = feelem_total + efgvol_total);
}


/* PARALLELIZATION FOR PVM  */

#define DCL_ACCESS_FUNC(TYPE, VAR)              \
  static TYPE                                   \
  get_##VAR(void) 

#define DEF_ACCESS_FUNC(TYPE, VAR)              \
  DCL_ACCESS_FUNC(TYPE, VAR)                    \
  {                                             \
    return VAR;                                 \
  }

#define DEF_VAR(TYPE, VAR, INITIALIZER)         \
  static TYPE VAR = INITIALIZER;                \
  DEF_ACCESS_FUNC(TYPE, VAR) 

static void
build_rep_parallel(hexmesh_model_t *model);
static void
build_partition(hexmesh_model_t *model, int i);
static void
print_partition_out_reqs(hexmesh_model_t *model, FILE *fp);
static void
print_partition_nodes(hexmesh_model_t *model, FILE *fp);
static void
print_partition_elems(hexmesh_model_t *model, FILE *fp);
static void
print_partition_conc_loads(hexmesh_model_t *model, FILE *fp);
static void
print_partition_mapping(hexmesh_model_t *model, FILE *fp);


static char file_name[256];
static char partition_file[256];

static char *
construct_partition_name(char *name, int i);


static char *
construct_partition_name(char *name, int i)
{
  sprintf(partition_file, "%s-w%d.in", name, i);
  return partition_file;
}

DCL_ACCESS_FUNC(int, total_if_nodes);
DCL_ACCESS_FUNC(int, num_partitions);
DCL_ACCESS_FUNC(int, partition_nodes);
DCL_ACCESS_FUNC(int, partition_elems);
DCL_ACCESS_FUNC(int, partition_loads);
DCL_ACCESS_FUNC(int, partition_total_if_nodes);


void
hexmesh_model_print_parallel(hexmesh_model_t *model, char *name)
{
  FILE *fp;
  int i;

  build_rep_parallel(model);

  sprintf(file_name, "%s-master.in", name);
  fp = fopen(file_name, "w");
  if (fp == NULL) {
    fprintf(stderr, "Failed fopen in %s (l. %d)\n", __FILE__, __LINE__);
    return;
  }

  if (model->description != NULL)
    fprintf(fp, "Master: %s\n", model->description);
  else
    fprintf(fp, "Master: no description\n");
  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "!     INPUT FILE FOR WASP; GENERATED BY HEXMESH \n");
  fprintf(fp, "!                (C) 1995 Petr Krysl \n");
  fprintf(fp, "! \n");

  fprintf(fp, "! #nodes #elements #materials #loads #time_funcs  \n");
  fprintf(fp, "   0         0         0        0        0   \n");

  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "PVM MASTER\n");
  fprintf(fp, "!  total # of interface nodes  \n");
  fprintf(fp, "   %d \n", get_total_if_nodes());
  fprintf(fp, "!  input file      [host name]   \n");
  for (i = 0; i < get_num_partitions(); i++) { 
    construct_partition_name(name, i);
    fprintf(fp, "\"%s\" ! HOST \n", partition_file);
  }
  fprintf(fp, "!  END PVM MASTER \n");
  fprintf(fp, "&\n");

  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "OPTIONS \n");
  fprintf(fp, "!  option_tag (O, T, S)     value \n");
  fprintf(fp, "!  END OPTIONS \n");
  fprintf(fp, "&\n");

  fprintf(fp, "!  END DATA \n");

  fclose(fp);

  for (i = 0; i < get_num_partitions(); i++) {
    construct_partition_name(name, i+1);
    fp = fopen(partition_file, "w");
    if (fp == NULL) {
      fprintf(stderr, "Failed fopen in %s (l. %d)\n", __FILE__, __LINE__);
      return;
    }
    build_partition(model, i+1);
    
    if (model->description != NULL)
      fprintf(fp, "Worker %d: %s\n", i, model->description);
    else
      fprintf(fp, "Worker %d: no description\n", i);
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "!     INPUT FILE FOR WASP; GENERATED BY HEXMESH \n");
    fprintf(fp, "!                (C) 1995 Petr Krysl \n");
    fprintf(fp, "! \n");
    
    fprintf(fp, "! #nodes #elements #materials #loads #time_funcs  \n");
    fprintf(fp, "   %d      %d        %d        %d        %d\n",
            get_partition_nodes(), get_partition_elems(),
            1, get_partition_loads(), 1);
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "NODES\n");
    fprintf(fp, "! id       x       y       z  CONSTRAIN: X  Y  Z (0=fix)\n");
    print_partition_nodes(model, fp);
    fprintf(fp, "!  END NODES \n");
    fprintf(fp, "&\n");
    
    fprintf(fp, "!  ======================================================\n");
    fprintf(fp, "ELEM\n");
    fprintf(fp, "   4  ! hexahedrons\n");
    fprintf(fp, "!  id    n1 n2 n3 n4     n5 n6 n7 n8    material \n");
    print_partition_elems(model, fp);
    fprintf(fp, "!  END ELEM \n");
    fprintf(fp, "&\n");
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "MATERIAL \n");
    fprintf(fp, "   5  \n");
    fprintf(fp, "!  id   E        nu    rho\n");
    fprintf(fp, "   1   %g   %g   %g \n", model->E, model->nu, model->rho);
    fprintf(fp, "!  END MATERIAL \n");
    fprintf(fp, "&\n");
    
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "TIME_FUNCTIONS \n");
    fprintf(fp, "  2  ! tabular \n");
    fprintf(fp, "! func#   #pairs \n");
    fprintf(fp, "   1        3\n");
    fprintf(fp, "! time     val\n");
    fprintf(fp, "   0        0\n");
    fprintf(fp, "   1        1\n");
    fprintf(fp, "   10000000 1\n");
    fprintf(fp, "!  END TIME_FUNCTIONS \n");
    fprintf(fp, "&\n");
    
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "LOADS \n");
    fprintf(fp, "! load type (1 = nodal) \n");
    fprintf(fp, "! node#  dir (0-5)  amplitude  #time_func \n");
    print_partition_conc_loads(model, fp);
    fprintf(fp, "!  END LOADS \n");
    fprintf(fp, "&\n");
    
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "! OUTPUT REQUESTS \n");
    fprintf(fp, "! vertex# \n");
    print_partition_out_reqs(model, fp);
    fprintf(fp, "!  END OUTPUT REQUESTS \n");
    
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "HISTORY \n");
    fprintf(fp, "!  client_tag (N, E, W)     entity# (0 for energy (W)) \n");
    fprintf(fp, "!  END HISTORY \n");
    fprintf(fp, "&\n");

    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "PVM WORKER\n");
    fprintf(fp, "!  total # of interface nodes on the worker's partition \n");
    fprintf(fp, "   %d \n", get_partition_total_if_nodes());
    fprintf(fp, "!  local number    global interface number \n");
    print_partition_mapping(model, fp);
    fprintf(fp, "!  END PVM WORKER \n");
    fprintf(fp, "&\n");
    
    fprintf(fp, "!  ===================================================== \n");
    fprintf(fp, "OPTIONS \n");
    fprintf(fp, "!  option_tag (O, T, S)     value \n");
    fprintf(fp, "!  END OPTIONS \n");
    fprintf(fp, "&\n");

    fclose(fp);
  } /* for (i = 0; i < num_of_partitions; i++) */

  fprintf(fp, "!  END DATA \n");
}



static void
find_if_nodes(hexmesh_model_t *model);
static void
find_number_of_partitions(hexmesh_model_t *model);

DEF_VAR(int, num_partitions, 0)
DEF_VAR(int, total_if_nodes, 0)
DEF_VAR(int, partition_total_if_nodes, 0)

static void
reset_conc_loads(hexmesh_model_t *model);


static void
build_rep_parallel(hexmesh_model_t *model)
{
  find_number_of_partitions(model);
  find_if_nodes(model);
  reset_conc_loads(model);
}

static void
reset_conc_loads(hexmesh_model_t *model)
{
  conc_load_t *cl;

  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    cl->included_in_output  = FALSE;
  }
}





static void
find_number_of_partitions(hexmesh_model_t *model)
{
  hexvol_t *e;
  
  num_partitions = 0;
  loop_through_list(model->hexvols, e, hexvol_t *) {
    num_partitions = max(num_partitions, e->partition);
  }
}
  


DEF_VAR(int, partition_nodes, 0)
DEF_VAR(int, partition_elems, 0)
DEF_VAR(int, partition_loads, 0)

static int curr_partition = 0;

#define NLoop() loop_through_list(model->nodes, n, node_t *) 
#define CLLoop() loop_through_list(model->conc_loads, cl, conc_load_t *)
#define ELoop() loop_through_list(model->hexvols, e, hexvol_t *) 
#define PRIN()                                                             \
  {                                                                        \
    NLoop() {                                                              \
      fprintf(stderr, "node %d, in_curr_partition %c, on_ghost_only %c\n", \
              n->fe_id, (n->in_curr_partition? 'T': 'F'),                  \
              (n->on_ghost_only? 'T': 'F'));                               \
    }                                                                      \
  }


static void
build_partition(hexmesh_model_t *model, int i)
{
  hexvol_t *e;
  node_t *n;
  conc_load_t *cl;
  int m;
#define IS_IN_P(e) (e->partition == curr_partition)
#define NUMEL(e)                                                        \
  { e->fe_id    = -1;  e->is_ghost = FALSE;                             \
    if (IS_IN_P(e)) {e->fe_id = ++partition_elems;}                     \
  }
#define ISIFN(e, m) (e->nodes[m]->iface_id > 0)
#define ISINCP(e, m) (e->nodes[m]->in_curr_partition)
#define INVEN(e) (e->fe_id < 1)
#define INVNN(n) (n->fe_id < 1)

  curr_partition = i;

  find_number_of_partitions(model);
  find_if_nodes(model);

  /* Mark all nodes as not being in the current partition */
  /* and invalidate their ids */
  NLoop(){n->in_curr_partition = FALSE;n->fe_id = -1;n->on_ghost_only = FALSE;}
  
  /* Number partition elements and */
  /* Find interface nodes which are in the current partition */
  /* Find partition total of interface nodes */
  partition_elems = 0;
  partition_total_if_nodes = 0;
  ELoop(){
    NUMEL(e);
    if (IS_IN_P(e))
      FOR(m, 8) {
        if (ISIFN(e, m)) {
          if (!e->nodes[m]->in_curr_partition)
            partition_total_if_nodes++;
          e->nodes[m]->in_curr_partition = TRUE;
        }
      }
  }
  /* PRIN(); */
  
  /* Add ghost elements */
  ELoop(){e->is_ghost = FALSE;}
  ELoop(){
    if (!IS_IN_P(e)) {
      FOR(m, 8) {
        if (ISIFN(e, m) && ISINCP(e, m)) {
          e->is_ghost = TRUE;
          if (INVEN(e))
            e->fe_id = ++partition_elems;
        }
      }
    }
  }

  /* Mark nodes which are not interface nodes, and are on ghosts  */
  NLoop(){ n->on_ghost_only = FALSE;}
  ELoop(){
    if (e->is_ghost) {
      FOR(m, 8){
        if (!ISIFN(e, m))
          e->nodes[m]->on_ghost_only = TRUE;
      }
    }
  }
  /* PRIN(); */

  /* Number local nodes */
  partition_nodes = 0;
  ELoop(){
    if (IS_IN_P(e) || e->is_ghost) {
      FOR(m, 8){
        if (INVNN(e->nodes[m]))
          e->nodes[m]->fe_id= ++partition_nodes;
        e->nodes[m]->in_curr_partition = TRUE;
      }
    }
  }
  /* PRIN(); */

  /* Partition loaded nodes */
  partition_loads = 0;
  CLLoop() {
    if (cl->node->in_curr_partition && (!cl->node->on_ghost_only)) {
      if (   cl->components.x != 0)
        partition_loads++;
      if (cl->components.y != 0)
        partition_loads++;
      if (cl->components.z != 0)
        partition_loads++;
    }
  }
}


static void
find_if_nodes(hexmesh_model_t *model)
{
  hexvol_t *e;
  node_t *n;
  int i;

  NLoop() { n->iface_id = 0; n->in_curr_partition = 0; }

  ELoop() {
    FOR(i, 8) 
      if (e->nodes[i]->in_curr_partition == 0)
        e->nodes[i]->in_curr_partition = e->partition;
  }
  
  total_if_nodes = 0;
  ELoop() {
    FOR(i, 8) {
      if (e->nodes[i]->in_curr_partition != e->partition)
        if (e->nodes[i]->iface_id < 1)
          e->nodes[i]->iface_id = ++total_if_nodes;
    }
  }
}



static void
print_partition_out_reqs(hexmesh_model_t *model, FILE *fp)
{
  output_request_t *or;

  loop_through_list(model->output_requests, or, output_request_t *) {
    if (or->node->in_curr_partition)
      fprintf(fp, "!   %d\n", or->node->fe_id);
  }
}


static void
print_partition_nodes(hexmesh_model_t *model, FILE *fp)
{
  node_t *n;

  loop_through_list(model->nodes, n, node_t *) {
    if (n->in_curr_partition)
      if (!n->on_ghost_only)
        fprintf(fp, "  %d   %g   %g  %g       %d   %d   %d\n",
                n->fe_id, n->loc.x, n->loc.y, n->loc.z, 
                n->constraint[0],
                n->constraint[1],
                n->constraint[2]);
      else
        fprintf(fp, "  %d   %g   %g  %g       %d   %d   %d\n",
                n->fe_id, n->loc.x, n->loc.y, n->loc.z, 
                CONSTRAINT_FIXED, CONSTRAINT_FIXED, CONSTRAINT_FIXED);
  }
}


static void
print_partition_elems(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *e;

  loop_through_list(model->hexvols, e, hexvol_t *) {
    if (e->partition == curr_partition || e->is_ghost) {
      fprintf(fp, " %d    %d %d %d %d     %d %d %d %d    1 ",
              e->fe_id,
              e->nodes[0]->fe_id,
              e->nodes[1]->fe_id,
              e->nodes[2]->fe_id,
              e->nodes[3]->fe_id,
              e->nodes[4]->fe_id,
              e->nodes[5]->fe_id,
              e->nodes[6]->fe_id,
              e->nodes[7]->fe_id);
      if (e->is_ghost)
        fprintf(fp, "  GHOST");
      fprintf(fp, "\n");
    }
  }
}


static void
print_partition_conc_loads(hexmesh_model_t *model, FILE *fp)
{
  conc_load_t *cl;

  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    if (   cl->node->in_curr_partition
        && (!cl->node->on_ghost_only)
        && (!cl->included_in_output)) {
      if (cl->components.x != 0) {
        fprintf(fp, "   1 \n");
        fprintf(fp, "  %d   DX   %g   %d\n",
                cl->node->fe_id, cl->components.x, 1);
      }
      if (cl->components.y != 0) {
        fprintf(fp, "   1 \n");
        fprintf(fp, "  %d   DY   %g   %d\n",
                cl->node->fe_id, cl->components.y, 1);
      }
      if (cl->components.z != 0) {
        fprintf(fp, "   1 \n");
        fprintf(fp, "  %d   DZ   %g   %d\n",
                cl->node->fe_id, cl->components.z, 1);
      }
      cl->included_in_output = TRUE;
    }
  }
}

  
static void
print_partition_mapping(hexmesh_model_t *model, FILE *fp)
{
  node_t *n;

  loop_through_list(model->nodes, n, node_t *) {
    if (n->iface_id > 0 && n->in_curr_partition) {
      fprintf(fp, "       %d             %d\n", n->fe_id, n->iface_id);
    }
  }
}



static void
print_tees_hcells(hexmesh_model_t *model, FILE *fp);
static void
print_tees_verts(hexmesh_model_t *model, FILE *fp, int symbolic,
                 int nodes_at_hex_centers);
static void
print_tees_concls(hexmesh_model_t *model, FILE *fp);
static void
print_tees_surf_tractions(hexmesh_model_t *model, FILE *fp,
                          int symbolic);
static void
print_tees_variables(hexmesh_model_t *model, FILE *fp);


void
hexmesh_model_print_TEES(hexmesh_model_t *model, FILE *fp,
                         int symbolic, int nodes_at_hex_centers)
{
  build_rep(model);
  
  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "!     INPUT FILE FOR TEES; GENERATED BY HEXMESH \n");
  fprintf(fp, "!                (C) 1995,1996 Petr Krysl \n");
  fprintf(fp, "! \n");

  if (model->description != NULL)
    fprintf(fp, "DESCRIPTION \"%s\"\n", model->description);
  else
    fprintf(fp, "DESCRIPTION unknown\n");

  fprintf(fp, "!  DOMAIN SETTINGS \n");
  fprintf(fp, "E %g\n", model->E);
  fprintf(fp, "NU %g\n", model->nu);
  fprintf(fp, "MAX_D_M %g\n", 2.0);
  fprintf(fp, "DEFAULT_D_M %g\n", 1.0);
  fprintf(fp, "INTG_ORDER 2\n");
  print_tees_variables(model, fp);
  fprintf(fp, "!  END DOMAIN SETTINGS \n");


  fprintf(fp, "! VERTICES =============================================== \n");
  fprintf(fp, "! V id x y z [NODE [EBC [X v][Y v][Z v]] [TYPE {EFG|FE}] [D_M v]]\n");
  print_tees_verts(model, fp, symbolic, nodes_at_hex_centers);
  fprintf(fp, "!  END VERTICES \n");

  fprintf(fp, "! HCELLS ================================================= \n");
  fprintf(fp, "! HCELL id n0 n1 n2 n3 n4 n5 n6 n7 [TYPE {EFG|FE}] [MAT mat]\n");
  print_tees_hcells(model, fp);
  fprintf(fp, "!  END HCELLS \n");

  fprintf(fp, "! CONCL ================================================== \n");
  fprintf(fp, "! CONCL id {X|Y|Z} amplitude vertex_id\n");
  print_tees_concls(model, fp);
  fprintf(fp, "!  END CONCL \n");

  fprintf(fp, "! HCELL SURFACE TRACTIONS ================================ \n");
  fprintf(fp, "! HST id  cell_id  face_id  tx0 ty0 ... t3z\n");
  print_tees_surf_tractions(model, fp, symbolic);
  fprintf(fp, "!  END HCELL SURFACE TRACTIONS \n");

  fprintf(fp, "!  END DATA \n");
}


static void
print_tees_variables(hexmesh_model_t *model, FILE *fp)
{
  char *name;
  double value;
  int stat;

  stat = expeval_first_var(&name, &value);
  while (stat != 0) {
    fprintf(fp, "EVAL \"%s = %g\"\n", name, value);
    stat = expeval_next_var(&name, &value);
  }
}


static void
print_tees_verts(hexmesh_model_t *model, FILE *fp,
                 int symbolic, int nodes_at_hex_centers)
{
  node_t *n;
  int id;

  id = 1;
  loop_through_list(model->nodes, n, node_t *) {
    n->efg_id = id++;
    fprintf(fp, "V  %d     %g %g %g ",
              n->efg_id, n->loc.x, n->loc.y, n->loc.z);
    if (   (!nodes_at_hex_centers)
        || (n->incident_to_iface_hexvol || n->incident_to_fe_hexvol)) {
      fprintf(fp, "NODE ");
      if (n->d_m > 0)
        fprintf(fp, "D_M %g ", n->d_m);
    }
    if (   n->type == FE_NODE
        || n->type == FE_EFG_NODE) {
      if (   n->constraint[0] == CONSTRAINT_FIXED
          || n->constraint[1] == CONSTRAINT_FIXED
          || n->constraint[2] == CONSTRAINT_FIXED) {
        fprintf(fp, "EBC ");
        if (n->constraint[0] == CONSTRAINT_FIXED)
          if (symbolic)
            fprintf(fp, "X \"%s\" ", n->symbolic_prescr_displ[0]);
          else
            fprintf(fp, "X %g ", n->prescr_displ[0]);
        if (n->constraint[1] == CONSTRAINT_FIXED)
          if (symbolic)
            fprintf(fp, "Y \"%s\" ", n->symbolic_prescr_displ[1]);
          else
            fprintf(fp, "Y %g ", n->prescr_displ[1]);
        if (n->constraint[2] == CONSTRAINT_FIXED)
          if (symbolic)
            fprintf(fp, "Z \"%s\" ", n->symbolic_prescr_displ[2]);
          else
            fprintf(fp, "Z %g ", n->prescr_displ[2]);
      }
      fprintf(fp, "TYPE FE ");
    } else {
      /* VOID */
    }
    fprintf(fp, "\n");
  }
  if (nodes_at_hex_centers) {
    hexvol_t *e;
    WCRec loc;
    
    loop_through_list(model->hexvols, e, hexvol_t *) {
      loc.x = (  e->nodes[0]->loc.x
               + e->nodes[1]->loc.x
               + e->nodes[2]->loc.x
               + e->nodes[3]->loc.x
               + e->nodes[4]->loc.x
               + e->nodes[5]->loc.x
               + e->nodes[6]->loc.x
               + e->nodes[7]->loc.x) / 8;
      loc.y = (  e->nodes[0]->loc.y
               + e->nodes[1]->loc.y
               + e->nodes[2]->loc.y
               + e->nodes[3]->loc.y
               + e->nodes[4]->loc.y
               + e->nodes[5]->loc.y
               + e->nodes[6]->loc.y
               + e->nodes[7]->loc.y) / 8;
      loc.z = (  e->nodes[0]->loc.z
               + e->nodes[1]->loc.z
               + e->nodes[2]->loc.z
               + e->nodes[3]->loc.z
               + e->nodes[4]->loc.z
               + e->nodes[5]->loc.z
               + e->nodes[6]->loc.z
               + e->nodes[7]->loc.z) / 8;
      fprintf(fp, "V  %d     %g %g %g ", id++, loc.x, loc.y, loc.z);
      fprintf(fp, "NODE ");
      if (e->d_m > 0)
        fprintf(fp, "D_M %g ", e->d_m);
      fprintf(fp, "\n");
    }
  }
}


static void
print_tees_hcells(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *e;
  int id;

  id = 1;
  loop_through_list(model->hexvols, e, hexvol_t *) {
    e->efg_id = id++;
    fprintf(fp, "HCELL %d    %d %d %d %d     %d %d %d %d ",
          e->efg_id,
          e->nodes[0]->efg_id,
          e->nodes[1]->efg_id,
          e->nodes[2]->efg_id,
          e->nodes[3]->efg_id,
          e->nodes[4]->efg_id,
          e->nodes[5]->efg_id,
          e->nodes[6]->efg_id,
          e->nodes[7]->efg_id);
    if (HV_IS_FE(e)) {
      fprintf(fp, "TYPE FE");
    } else {
      fprintf(fp, "TYPE EFG");
    }
    if (e->quadrature_order > 0)
      fprintf(fp, " QR %d ", e->quadrature_order);
    fprintf(fp, "\n");
  }
}


static void
print_tees_concls(hexmesh_model_t *model, FILE *fp)
{
  conc_load_t *cl;
  int id;

  id = 0;
  loop_through_list(model->conc_loads, cl, conc_load_t *) {
    id++;
    if (cl->components.x != 0) {
      fprintf(fp, "CONCL  %d   X   %g   %d\n",
              id, cl->components.x, cl->node->efg_id);
    }
    if (cl->components.y != 0) {
      fprintf(fp, "CONCL  %d   Y   %g   %d\n",
              id, cl->components.y, cl->node->efg_id);
    }
    if (cl->components.z != 0) {
      fprintf(fp, "CONCL  %d   Z   %g   %d\n",
              id, cl->components.z, cl->node->efg_id);
    }
  }
}


static void
print_tees_surf_tractions(hexmesh_model_t *model, FILE *fp,
                          int symbolic)
{
  hexvol_t *h;
  int id, f, n;

  id = 0;
  loop_through_list(model->hexvols, h, hexvol_t *) {
    FOR(f, 6) {
      if (h->face_traction[f] != NULL) {
        fprintf(fp,
                "HST %d   %d   %d   ", ++id, h->efg_id, f);
        if (symbolic) {
          fprintf(fp, "SYMBOLIC \"%s\" \"%s\" \"%s\"",
                  h->face_traction[f]->symbolic_traction[0],
                  h->face_traction[f]->symbolic_traction[1],
                  h->face_traction[f]->symbolic_traction[2]);
        } else {
          FOR(n, 4) {
            fprintf(fp, "%g %g %g    ",
                    h->face_traction[f]->node_traction[n].x,
                    h->face_traction[f]->node_traction[n].y,
                    h->face_traction[f]->node_traction[n].z);
          }
        }
        fprintf(fp, "\n");
      } 
    }
  }
}





static void
print_ndot_hcells(hexmesh_model_t *model, FILE *fp);
static void
print_ndot_verts(hexmesh_model_t *model, FILE *fp);
static void
print_ndot_surf_tractions(hexmesh_model_t *model, FILE *fp);


void
hexmesh_model_print_NDOT(hexmesh_model_t *model, FILE *fp)
{
  build_rep(model);
  
  fprintf(fp, "!  ======================================================= \n");
  fprintf(fp, "!     INPUT FILE FOR NDOT;  GENERATED BY HEXMESH \n");
  fprintf(fp, "!                (C) 1995,1996 Petr Krysl \n");
  fprintf(fp, "! \n");

  if (model->description != NULL)
    fprintf(fp, "! Model: \"%s\"\n", model->description);
  else
    fprintf(fp, "! Model:  unknown description\n");

  fprintf(fp, "Constants \n");
  fprintf(fp, "lambda = %g\n",
          model->E * model->nu / (1 + model->nu) / (1 - 2 * model->nu));
  fprintf(fp, "mu = %g\n", model->E / 2 / (1 + model->nu));
  fprintf(fp, "d_max = %g\n", 2.0);
  fprintf(fp, "&\n");

  fprintf(fp, "Nodes\n");
  fprintf(fp, "! node # x y z  EFG/FE fix_x fix_y fix_z u_x u_y u_z\n");
  print_ndot_verts(model, fp);
  fprintf(fp, "&\n");

  fprintf(fp, "Integration cells\n");
  fprintf(fp, "! cell #  n0 n1 n2 n3 n4 n5 n6 n7  q_order\n");
  print_ndot_hcells(model, fp);
  fprintf(fp, "&\n");

  fprintf(fp, "Tractions\n");
  fprintf(fp, "! traction #  n0 tx0 ty0 tz0 ... n3 tx3 ty3 t3z\n");
  print_ndot_surf_tractions(model, fp);
  fprintf(fp, "&\n");

  fprintf(fp, "!  END OF DATA \n");
}




static void
print_ndot_verts(hexmesh_model_t *model, FILE *fp)
{
  node_t *n;
  int id;

  id = 1;
  loop_through_list(model->nodes, n, node_t *) {
    n->efg_id = id++;
    fprintf(fp, " %8d   %g %g %g    %d    ",
            n->efg_id, n->loc.x, n->loc.y, n->loc.z,
            (n->type == FE_NODE || n->type == FE_EFG_NODE? 1: 0));
    if (   n->type == FE_NODE
        || n->type == FE_EFG_NODE) {
      if (   n->constraint[0] == CONSTRAINT_FIXED
          || n->constraint[1] == CONSTRAINT_FIXED
          || n->constraint[2] == CONSTRAINT_FIXED) {
        if (n->constraint[0] == CONSTRAINT_FIXED)
          fprintf(fp, "%d ", CONSTRAINT_FIXED);
        else
          fprintf(fp, "%d ", CONSTRAINT_FREE);
        if (n->constraint[1] == CONSTRAINT_FIXED)
          fprintf(fp, "%d ", CONSTRAINT_FIXED);
        else
          fprintf(fp, "%d ", CONSTRAINT_FREE);
        if (n->constraint[2] == CONSTRAINT_FIXED)
          fprintf(fp, "%d ", CONSTRAINT_FIXED);
        else
          fprintf(fp, "%d ", CONSTRAINT_FREE);
        fprintf(fp, " %g %g %g ",
                n->prescr_displ[0], n->prescr_displ[1], n->prescr_displ[2]);
      } else {
        fprintf(fp, "%d %d %d ",
                CONSTRAINT_FREE, CONSTRAINT_FREE, CONSTRAINT_FREE);
        fprintf(fp, " %g %g %g ", 0.0, 0.0, 0.0);
      }
    } else {
    }
    fprintf(fp, "\n");
  }
}


static void
print_ndot_hcells(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *e;
  int id;

  id = 1;
  loop_through_list(model->hexvols, e, hexvol_t *) {
    e->efg_id = id++;
    fprintf(fp, " %8d   %d %d %d %d   %d %d %d %d    %d",
          e->efg_id,
          e->nodes[0]->efg_id,
          e->nodes[1]->efg_id,
          e->nodes[2]->efg_id,
          e->nodes[3]->efg_id,
          e->nodes[4]->efg_id,
          e->nodes[5]->efg_id,
          e->nodes[6]->efg_id,
          e->nodes[7]->efg_id,
            model->quadrature_order);
    fprintf(fp, "\n");
  }
}



static void
print_ndot_surf_tractions(hexmesh_model_t *model, FILE *fp)
{
  hexvol_t *h;
  int id, f, n;

  id = 0;
  loop_through_list(model->hexvols, h, hexvol_t *) {
    FOR(f, 6) {
      if (h->face_traction[f] != NULL) {
        FOR(n, 4) {
          fprintf(fp, "%d %g %g %g    ",
                  h->nodes[face_vertex_indices[f][n]]->efg_id, 
                  h->face_traction[f]->node_traction[n].x,
                  h->face_traction[f]->node_traction[n].y,
                  h->face_traction[f]->node_traction[n].z);
        }
        fprintf(fp, "\n");
      } 
    }
  }
}


void
compute_d_ms(hexmesh_model_t *model, double diag_length_mult);

void
hexmesh_compute_d_m(hexmesh_model_t *model, double diag_length_mult)
{
  compute_d_ms(model, diag_length_mult);
}


#ifdef HAVE_RANDOM
#   define RANDOM_RETURNS_MAX LONG_MAX
#else  /* !AIXV3 */
#   ifdef HAVE_RAND
#      include <stdlib.h>
#      define random() rand()
#      define srandom(_seed) srand(_seed)
#      define RANDOM_RETURNS_MAX 32767
#   else /* everything else */
       I am really bad off without any random generator;
#      define RANDOM_RETURNS_MAX LONG_MAX
#   endif
#endif
static int
one_of_four(void);
static int
flip(int two);


static int
one_of_four(void)
{
  
  if (flip(2))
    if (flip(2))
      return 1;
    else
      return 2;
  else
    if (flip(2))
      return 3;
    else
      return 4;
}


static int
flip(int two)
{
  static int seeded = 0;
  long rn;

  if (!seeded) { srandom(131313); seeded = 1; }

  rn = random();

  return (((rn / two) * 2) == rn);
}

static void
rotate_diagonal(hexvol_t *h);

void
hexmesh_randomize_06_diagonal(hexmesh_model_t *model)
{
  hexvol_t *h;
  loop_through_list(model->hexvols, h, hexvol_t *) { rotate_diagonal(h); }
}

static void
draw_diagonal(hexvol_t *h);

static void
rotate_diagonal(hexvol_t *h)
{
  int which = one_of_four();
#define FOUR_SHIFT(start)                       \
  {                                             \
    node_t *t = h->nodes[start];                \
    h->nodes[start  ] = h->nodes[start+1];      \
    h->nodes[start+1] = h->nodes[start+2];      \
    h->nodes[start+2] = h->nodes[start+3];      \
    h->nodes[start+3] = t;                      \
  }
#define SHIFT_ONCE() { FOUR_SHIFT(0); FOUR_SHIFT(4); }

  switch (which) {
  case 1:
    /* do nothing */
    break;
  case 2:
    /* rotate once */
    SHIFT_ONCE();
    break;
  case 3:
    /* rotate twice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  case 4:
    /* rotate thrice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  default:
    break;
  }
  draw_diagonal(h);
}


static GraphicObj *diagonal = NULL;

static void
draw_diagonal(hexvol_t *h)
{
  WCRec pts[2] = {{0,0,0}, {0,0,0}};

  if (diagonal == NULL) { diagonal = CreateLine3D(pts);
                          EGWithMaskChangeAttributes(COLOR_MASK, diagonal);
                        }
  pts[0] = h->nodes[0]->loc;
  pts[1] = h->nodes[6]->loc;
  EGModifyGraphicsGeometry(diagonal, pts);
  EMDrawGraphics(ESIModel(), diagonal);
}
